"""
This module includes all the functionalities for epipolar geometry.

"""

__author__ = 'Christian Pieringer'
__created__ = '2019.10.08'

import numpy as np


def line2points(x, l):
    """
    Transforms a point into a line.

    Args:
        x: array with a point
        l: array with a line in homogeneous coordinates
    
    Returns:
        an array with the line
    """
    y = (-1.0 / l[1]) * (l[2] - l[0] * x)

    return np.array([x, y])


def plot_epipolar_line(mat_f, m, ax, line_color='r'):
    """
    This method plot an epipolar line into an open figure.

    Args:
        mat_f: numpy array with the fundamental matrix
        m: numpy array with a image point (2D)
        ax: a matplotlib figure axe
        line_color: an string indicating the line color
    
    Returns:
        ax: the figure axe

    """
    if m.shape[-1] == 2:
        m = np.hstack([m, np.ones((m.shape[0], 1))])

    ell = np.dot(mat_f, m)
    ell = ell / ell[-1]
    ax_lim = ax.axis()
    x = np.asarray(ax_lim[0:2])
    a = ell[0]
    b = ell[1]
    c = ell[2]

    y = -(c + a * x) / b

    ax.plot(x, y, line_color)
    ax.axis('off')

    return ax


def estimate_trifocal_tensor(mat_a, mat_b, mat_c):
    """

    :param mat_a:
    :param mat_b:
    :param mat_c:
    :return:
    """

    mat_h = np.vstack([mat_a, np.random.rand(1, 4)])
    while np.abs(np.linalg.det(mat_h)) < 0.001:
        mat_h = np.vstack([mat_a, np.random.rand(1, 4)])
        if np.linalg.det(mat_h) < 0.001:
            print('Warning: determinant of H in trifocal estimation is zero')

    a = np.dot(mat_a, np.linalg.pinv(mat_h))
    b = np.dot(mat_b, np.linalg.pinv(mat_h))  # b = B * inv(H)
    c = np.dot(mat_c, np.linalg.pinv(mat_h))  # c = C * inv(H)

    tt = np.zeros((3, 3, 3))

    for i in range(3):
        for j in range(3):
            for k in range(3):
                tt[i, j, k] = b[j, i] * c[k, 3] - b[j, 3] * c[k, i]

    #     for i in range(3):
    #         t1 = np.dot(b[:, i].reshape(-1, 1), c[:, -1].reshape(1, -1))
    #         t2 = np.dot(b[:, -1].reshape(-1, 1), c[:, i].reshape(1, -1))
    #         tt[i, :, :] = t1 - t2
    return tt


def estimate_fundamental_matrix(mat_a, mat_b, method='pseudo'):
    """
    This method computes the 3x3 fundamental matrix from 3x4 projection matrices A and B according methods defined in
    method parameter. We refer the reader to [1] for more details about this implementation.

    Args:
        mat_a: numpy array with projection matrix of view 1
        mat_b: numpy array with projection matrix of view 2
        method: a string defining computation method. If no method is given,
            'pseudo' will be assumed as default.
            'tensor' : uses bifocal tensors with canonic matrix
            'tensor0': uses bifocal tensors without canonic matrix
            'pseudo' : uses pseudoinverse matrix of A (default)

    Returns:
        mat_f: the 3x3 fundamental matrix.

    [1] R. Hartley and A. Zisserman. Multiple View Geometry in Computer Vision.
    Cambridge University Press, 2000.
    """

    if method.lower() == 'tensor':
        mat_h = np.vstack([mat_a, np.random.rand(4)])

        while np.abs(np.linalg.det(mat_h)) < 0.001:
            mat_h = np.vstack([mat_a, np.random.rand(4)])

        mat_bs = np.dot(mat_b, np.linalg.inv(mat_h))
        b = np.vstack([mat_bs, mat_bs])

        mat_f = np.zeros((3, 3))

        for i in range(3):
            for j in range(3):
                mat_f[i, j] = b[i + 1, j] * b[i + 2, 3] - b[i + 2, j] * b[
                    i + 1, 3]

    elif method.lower() == 'tensor0':
        # TODO: Verificar con Domingo
        mat_f = np.zeros((3, 3))

        for i in range(3):
            sb = mat_b.copy()
            sb = np.delete(sb, i, 0)

            for j in range(3):
                sa = mat_a.copy()
                sa = np.delete(sa, j, 0)

                mat_f[i][j] = (-1) ** (i + j) * np.linalg.det(
                    np.vstack([sa, sb]))

    else:
        # Find mat_c1 as the null space of the projection matrix using SVD factorization of mat_a. The null vector is 
        # the last row in the matrix V.T . See numpy documentation for more details.

        _, _, mat_v = np.linalg.svd(mat_a)
        mat_c1 = mat_v[-1, :].reshape(-1, 1)
        mat_u = skew_symmetric(mat_b, mat_c1)

        # Compute the fundamental matrix
        mat_f = np.asmatrix(mat_u) * np.asmatrix(mat_b) * np.asmatrix(
            np.linalg.pinv(mat_a))
        mat_f = np.array(mat_f)

    mat_f = np.array(mat_f / np.linalg.norm(mat_f)[0])  # Normalize as F / norm(F)

    return mat_f


def skew_symmetric(mat, null_space):
    """
    Compute the skew-symmetric matrix.

    :param mat: numpy array with a projection matrix (3x4)
    :param null_space: an array with the null space
    :return: an array with the skew-symmetric matrix.

    """
    # epipole in second view
    e2 = np.dot(mat, null_space)

    # The vector e2 as Skew-symmetric matrix
    mat_u = np.array(
        [
            [0, -e2[2], e2[1]],
            [e2[2], 0, -e2[0]],
            [-e2[1], e2[0], 0]
        ]
    )

    return mat_u


def reproject_trifocal(m1, m2, tensor, method=2):
    """

    Args:
        m1: 2D point in image 1
        m2: 2D point in image 2
        tensor: 
        method:
    
    Returns:
        m3: a 2D points in image 3
    """

    # Assert that m1 and m2 are homogeneous.
    m1 = m1 / m1[-1, :]
    m2 = m2 / m2[-1, :]

    n = m1.shape[-1]
    m3 = np.zeros((3, n))

    for k in range(n):
        temp = tensor[:, 0, :] - m2[0, k] * tensor[:, 2, :]
        m3[:, k] = (temp.T @ m1[:, k][:, np.newaxis]).ravel()

    #         temp = np.array([
    #             tensor[:, 0, 0] - m2[0, k] * tensor[:, 2, 0],
    #             tensor[:, 0, 1] - m2[0, k] * tensor[:, 2, 1],
    #             tensor[:, 0, 2] - m2[0, k] * tensor[:, 2, 2]
    #         ]) @ m1[:, k][:, np.newaxis]

    #         m3[:, k] = temp.ravel()

    m3 = m3 / m3[-1, :]

    if method == 2:
        m32 = np.zeros((3, n))
        for k in range(n):
            temp = tensor[:, 1, :] - m2[1, k] * tensor[:, 2, :]
            m32[:, k] = (temp.T @ m1[:, k][:, np.newaxis]).ravel()

        #             temp = np.array(
        #                 [
        #                     tensor[:, 1, 0] - m2[1, k] * tensor[:, 2, 0],
        #                     tensor[:, 1, 1] - m2[1, k] * tensor[:, 2, 1],
        #                     tensor[:, 1, 2] - m2[1, k] * tensor[:, 2, 2]
        #                 ]) @ m1[:, k][:, np.newaxis]#reshape(-1, 1)).ravel()
        #             m32[:, k] = temp.ravel()
        m32 = m32 / m32[-1, :]
        m3 = (m3 + m32) / 2

    return m3


def recon_3dn(m, mat_p, method=None):
    """
    This method computes the reconstruction of a M point from two dimensional
    points in images.

    Args:
        m:
        mat_p:
        method:
        ls: the function computes the reconstruction using least-square method
    
    Returns:
        mat_mr: Reprojected points
        ms: Estimated points in 2D
        err: reprojection error in pixels
    """

    # n = len(m)  # number of images
    # mat_q = []
    # mat_r = []
    #
    # for xy, p in zip(m, mat_p):
    #     mat_q.append([p[2, 0] * xy[0] - p[0, 0], p[2, 1] * xy[0] - p[0, 1],
    #                   p[2, 2] * xy[0] - p[0, 2]])
    #     mat_q.append([p[2, 0] * xy[1] - p[1, 0], p[2, 1] * xy[1] - p[1, 1],
    #                   p[2, 2] * xy[1] - p[1, 2]])
    #     mat_r.append([p[0, 3] - p[2, 3] * xy[0]])
    #     mat_r.append([p[1, 3] - p[2, 3] * xy[1]])
    #
    # mat_q = np.array(mat_q)
    # mat_r = np.array(mat_r)
    #
    # print(mat_q.shape)
    # print(mat_r.shape)

    n = m.shape[-1]

    mat_q = np.zeros((2 * n, 3))
    mat_r = np.zeros((2 * n, 1))

    for k in range(n):
        x = m[0, k] / m[2, k]
        y = m[1, k] / m[2, k]
        p = mat_p[k * 3:k * 3 + 3, :]
        mat_q[k * 2:k * 2 + 2, :] = np.array(
            [
                [p[2, 0] * x - p[0, 0], p[2, 1] * x - p[0, 1],
                 p[2, 2] * x - p[0, 2]],
                [p[2, 0] * y - p[1, 0], p[2, 1] * y - p[1, 1],
                 p[2, 2] * y - p[1, 2]]
            ]
        )

        mat_r[k * 2:k * 2 + 2, :] = np.array(
            [
                [p[0, 3] - p[2, 3] * x],
                [p[1, 3] - p[2, 3] * y]
            ]
        )

    mat_mr = np.vstack([np.dot(np.linalg.pinv(mat_q), mat_r), 1.0])  # Compute [Q.T * Q] * Q.T as pseudoinverse
    
    ms = mat_p @ mat_mr
    ms = np.reshape(ms, (n, 3)).T
    ms = ms / ms[-1, :]
    d = ms[0:1, :] - m[0:1, :]
    err = np.sqrt(np.sum(d * d, 1)).item()

    return mat_mr, ms, err



import numpy as np
import cv2
from scipy.ndimage.morphology import binary_dilation as imdilate
def superimpose_xray_images(Ib,If,i,j):

    B = 0.1
    C = 230 
    
    It = Ib.copy()

    i0b = max(0,i)
    i1b = min( max(0, i + If.shape[0]), Ib.shape[0])
    j0b = max(0,j)
    j1b = min( max(0, j + If.shape[1]), Ib.shape[1])
    Jb = (Ib[i0b:i1b,j0b:j1b]-B)/C

    i0f =  max(-i, 0)
    i1f = min(Ib.shape[0] - i, If.shape[0])
    j0f = max(-j, 0)
    j1f = min(Ib.shape[1] - j, If.shape[1])
    Jf = (If[i0f: i1f,j0f  : j1f]- B)/C
    
    It[i0b:i1b,j0b:j1b] = C*Jf*Jb + B

    return It


def draw_bounding_box(I,y1,x1,dy,dx,label):
    # Modified from  Project: Traffic_sign_detection_YOLO   
    # Author: AmeyaWagh   
    # File: objectDetectorYOLO.py    
    # License: MIT 
    y2 = y1+dy
    x2 = x1+dx
    cv2.rectangle(I,(x1,y1),(x2,y2),(0,255,0),6)
    font = cv2.FONT_HERSHEY_PLAIN
    labelSize=cv2.getTextSize(label,font,0.5,2)
    _x1 = x1
    _y1 = y1
    _x2 = _x1+labelSize[0][0]*12
    _y2 = y1-int(labelSize[0][1])*12
    cv2.rectangle(I,(_x1,_y1),(_x2,_y2),(0,255,0),cv2.FILLED)
    cv2.putText(I,label,(x1-5,y1-5),font,5.5,(200,200,200),3,cv2.LINE_AA)
    return I


def mask_simulation(I,h,i0,j0):
    J = I.copy()
    (mi,mj) = h.shape
    mi2  = int(np.fix(mi/2))
    mj2  = int(np.fix(mj/2))
    i1 = i0-mi2 
    i2 = i0+mi2+1
    j1 = j0-mj2 
    j2 = j0+mj2+1
    J[i1:i2,j1:j2] = I[i1:i2,j1:j2]*(1+h)
    return   J

def funcf(m,K):
    x = np.ones((3,1))
    x[0] = m[0]
    x[1] = m[1]
    y = np.matmul(K,x)
    return y


def ellipsoid_simulation(I,K,SSe,f,abc,var_mu,xmax):

    J = I.copy()
    (N,M) = I.shape
    

    R = np.zeros((N,M)) # ROI of simulated defect

    invK  = np.linalg.inv(K)

    if len(abc)==3: # elliposoid
        (a,b,c) = abc 
    else: # sphere
        a = abc 
        b = a
        c = a 

    # Computation of the 3 x 3 matrices Phi and L
    H     = np.linalg.inv(SSe)
    h0    = H[0,:]/a
    h1    = H[1,:]/b
    h2    = H[2,:]/c
    Hs    = np.zeros((3,3))
    Hs[:,0] = h0[0:3]
    Hs[:,1] = h1[0:3]
    Hs[:,2] = h2[0:3]
    hd      = np.zeros((3,1))
    hd[0] = h0[3]
    hd[1] = h1[3]
    hd[2] = h2[3]
    Phi = np.matmul(Hs,Hs.T)
    hhd = np.matmul(hd,hd.T)
    hhd1 = 1-np.matmul(hd.T,hd)
    L = np.matmul(np.matmul(Hs,hhd),Hs.T) + hhd1*Phi

    # Location of the superimposed area
    A     = L[0:2,0:2]
    mc    = np.array(-f*np.matmul(np.linalg.inv(A),L[0:2,2]))
    x     = np.linalg.eig(A)[0]
    C     = np.array([x[1],x[0]])
    la    = C
    a00   = np.linalg.det(L)/np.linalg.det(A)
    ae    = f*np.sqrt(-a00/la[0])
    be    = f*np.sqrt(-a00/la[1])
    al    = np.arctan2(C[1],C[0])+np.pi
    ra    = np.array( [ae*np.cos(al), ae*np.sin(al)] )
    rb    = np.array( [be*np.cos(al+np.pi/2), be*np.sin(al+np.pi/2)] )
    u1    = funcf(mc+ra,K)
    u2    = funcf(mc+rb,K)
    u3    = funcf(mc-ra,K)
    u4    = funcf(mc-rb,K)
    uc    = funcf(mc,K)
    e1    = u1+u2-uc
    e2    = u1+u4-uc
    e3    = u3+u2-uc
    e4    = u3+u4-uc
    E     = np.concatenate((e1,e2,e3,e4),axis=1)
    Emax  = np.max(E,axis=1)
    Emin  = np.min(E,axis=1)
    umin  = int(max(np.fix(Emin[0]), 1))
    vmin  = int(max(np.fix(Emin[1]), 1))
    umax  = int(min(np.fix(Emax[0]+1), N))
    vmax  = int(min(np.fix(Emax[1]+1), M))
    q  = 255/(1-np.exp(var_mu*xmax))
    R[umin:umax,vmin:vmax] = 1
    R = imdilate(R)
    z = np.zeros((2,1))
    for u in range(umin,umax):
        z[0] = u
        for v in range(vmin,vmax):
            z[1] = v
            m = funcf(z,invK)
            m[0:2] = m[0:2]/f
            p = np.matmul(np.matmul(m.T,L),m)
            if p>0:
                x = np.matmul(np.matmul(m.T,Phi),m)
                d = 2*np.sqrt(p)*np.linalg.norm(m)/x
                J[u,v] = np.exp(var_mu*d)*(I[u,v]-q)+q

    return J


def voxels_simulation(N,M,V,s,P):
    Q = np.zeros((N,M))
    (Nx,Ny,Nz) = V.shape
    for x in range(Nx):
        print(str(x)+'/'+str(Nx))
        X = x/s
        for y in range(Ny):
            Y = y/s
            for z in range(Nz):
                if V[x,y,z]:
                    Z = z/s
                    Mp = np.array([X,Y,Z,1])
                    w = np.matmul(P,Mp)
                    w = w/w[2]
                    i1 = int(np.fix(w[0]))
                    if i1 >= 0 and i1<N:
                        j1 = int(np.fix(w[1]))
                        if j1 >= 0 and j1<M:
                            i2 = i1+1
                            j2 = j1+1
                            a1 = w[0]-i1     
                            b1 = w[1]-j1
                            a2 = 1-a1          
                            b2 = 1-b1
                            Q[i1,j1] = Q[i1,j1] + a2*b2
                            Q[i1,j2] = Q[i1,j2] + a2*b1
                            Q[i2,j1] = Q[i2,j1] + a1*b2
                            Q[i2,j2] = Q[i2,j2] + a1*a1

    return Q



import numpy as np
import matplotlib.pyplot as plt
from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.plotting import plot_sequential_feature_selection as plot_sfs

from sklearn.feature_selection import RFE, RFECV
from sklearn.svm import SVR
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import Ridge
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier
from sklearn.naive_bayes import GaussianNB
from pybalu.feature_selection import clean, exsearch
from pybalu.feature_transformation import normalize, pca
from pybalu.feature_selection import sfs as sfsfisher


def defineModel(bcl):
    model = eval(bcl[0]+'('+bcl[1]+')')
    return model

def fse_sbs(bcl,X,d,m):
    estimator = defineModel(bcl)
    selector  = RFE(estimator, m, step=1, verbose = 2)
    selector  = selector.fit(X, d)
    sel       = np.nonzero(selector.support_)[0]
    return sel


def fse_sbs_cv(bcl,X,d,m):
    estimator = defineModel(bcl)
    selector  = RFECV(estimator, min_features_to_select=m,step=1, cv=5)
    selector  = selector.fit(X, d)
    print(selector.grid_scores_)
    sel       = np.nonzero(selector.support_)[0]
    plt.figure()
    plt.xlabel("Number of features selected")
    plt.ylabel("Cross validation score (nb of correct classifications)")
    plt.plot(range(1, len(selector.grid_scores_) + 1), selector.grid_scores_)
    plt.show()    
    return sel




def fse_sfs(bcl,X,d,m,cv = 0, show = 0):
    estimator = defineModel(bcl)
    sfs = SFS(estimator, k_features=m,forward=True,floating=False,verbose=2,scoring='accuracy',cv=cv)
    sfs = sfs.fit(X, d)
    sel = sfs.k_feature_idx_
    print(' ')
    if show:
        plot_sfs(sfs.get_metric_dict(), kind='std_err')
        plt.title('Sequential Forward Selection (w. StdErr)')
        plt.grid()
        plt.show()
    return sel


def fsel(bcl,X,d,m,forward=True,floating=False,cv = 0, show = 0):
    if show>0:
        print('Feature Selection - '+bcl[0]+':  - number of features reducing from ' + str(X.shape[1]) + ' to '+str(m)+' ...')
    if bcl[0]=='Fisher':
        sel = sfsfisher(X,d,m)
    else:
        estimator = defineModel(bcl)
        sfs = SFS(estimator, k_features=m,forward=True,floating=False,verbose=show,scoring='accuracy',cv=cv)
        sfs = sfs.fit(X, d)
        sel = list(sfs.k_feature_idx_)
        if show>0:
            print(' ')
        if show:
            plot_sfs(sfs.get_metric_dict(), kind='std_err')
            plt.title('Sequential Forward Selection')
            plt.grid()
            plt.show()
    return sel

def sfspca_old(X,d,m1,m2,cc,ex,m3):
    s1 = sfsfisher(X,d,m1)
    X  = X[:,s1]
    Y, _, _, _, _ = pca(X, n_components=m2)
    if cc == 1:
        Y = np.concatenate((X,Y),axis=1)
    if ex == 1:
        s2 = exsearch(Y,d,m3)
    else:
        s2 = sfs(Y,d,m3)
    X = Y[:,s2]
    return X

def sfspca(bcl,X,d,m1,m2,cc,ex,m3):
    s1 = fsel(bcl,X,d,m1)
    X  = X[:,s1]
    if m2>0:
        Y, _, _, _, _ = pca(X, n_components=m2)
    else:
        Y = X
    if cc == 1:
        Y = np.concatenate((X,Y),axis=1)
    if m3 > 0:
        if ex == 1:
            s2 = exsearch(Y,d,m3)
        else:
            s2 = fsel(bcl,Y,d,m3)
        X = Y[:,s2]
    else:
        X = Y
    return X


def clean_norm(X):
    sclean        = clean(X,show=True)
    X             = X[:,sclean]
    X,a,b         = normalize(X)
    return X,sclean,a,b

def clean_norm_transform(Xt,sclean,a,b):
    Xt            = Xt[:,sclean]
    Xt            = Xt*a + b
    return Xt




def fse_model(sto):
    st = sto.lower()
    if st == 'lr':
        name    = 'LogisticRegression'            # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'ridge':
        name    = 'Ridge'                         # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'lda':
        name    = 'LinearDiscriminantAnalysis'    # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'qda':
        name    = 'QuadraticDiscriminantAnalysis' # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'svm-lin':
        name    = 'SVC'                           # name of the classifier
        params  = 'kernel="linear"'               # parameters of the classifier
    elif st == 'svm-rbf':
        name = 'SVC'
        params = 'kernel = "rbf", gamma=0.01, C=0.01'
    elif st == 'bayes-naive':
        name = 'GaussianNB'
        params = ''
    elif st == 'nn':
        name = 'MLPClassifier'
        params = 'solver="adam", alpha=1e-5,hidden_layer_sizes=(10,), random_state=1,max_iter=2000'
    elif st == 'rf':
        name = 'RandomForestClassifier'
        params = 'n_estimators=20,random_state = 0'
    elif st == 'fisher':
        name = 'Fisher'
        params = ''
    return name,params





"""
This module contains a collection of special functions use generally in 
2D filtering task.

"""

__author__ = 'Christian Pieringer'
__created__ = '2020.01.14'

import numpy as np


def prewitt_kernel(size=3, *args):
    """
    This function create a Prewitt Kernel. The default size is 3 x 3 kernel.
    :param size:
    :param args:
    :return:
    """

    prewitt = np.array([[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]], dtype='double')
    return prewitt


def sobel_kernel(size=3, *args):
    """
    This method create a Sobel Kernel. The standar size is a 3 x 3 kernel.

    :param size: int value to define the kernel size.

    :return:
    """

    sobel = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]], dtype='double')
    return np.transpose(sobel)


def gaussian_kernel(size, sigma, *args):
    """
    This method creates a symmetric Gaussian kernel.

    :param sigma: float number that defines the width of the Gaussian
            distribution
    :param size: inf number to specify the size of the kernel.
            Typical numbers 3, 5, 9, 11.
    """
    x, y = np.mgrid[-size // 2 + 1:size // 2 + 1,
           -size // 2 + 1:size // 2 + 1]

    kernel = np.exp(-((x ** 2 + y ** 2) / (2.0 * sigma ** 2)))
    kernel = kernel / kernel.sum()

    return kernel.astype('double')


def log_kernel(size, sigma, *args):
    """
    This option creates a Laplacian of Gaussian Kernel centered at 
    the middle of the kernel

    :param size: integer value to define the size of the kernel.
    :param sigma: float value for the scale factor.
    :returns: a numpy array with the LoG kernel.
    
    """

    x, y = np.meshgrid(np.arange(-size - 1 / 2 + 1, size - 1 / 2 + 1),
                       np.arange(-size - 1 / 2 + 1, size - 1 / 2 + 1))

    # LoG filter
    # normal = -1.0 / (2.0 * np.pi * sigma ** 2)

    # kernel = ((x ** 2 + y ** 2 - (2.0 * sigma ** 2)) / sigma ** 4)
    # kernel = kernel * np.exp(-(x ** 2 + y ** 2) / (2.0 * sigma ** 2))  
    # kernel = kernel / normal

    kernel = x ** 2 + y ** 2
    kernel = (-1 / (np.pi * sigma ** 4)) * (
                1 - kernel / (2 * sigma ** 2)) * np.exp(
        -kernel / (2 * sigma ** 2))

    # Normalize ensuring 0 DC. See Peter Kovesi's Toolbox
    kernel = kernel - np.mean(kernel.flatten())

    return kernel



import numpy as np
import pandas as pd
import cv2
import matplotlib.pyplot as plt
from scipy.ndimage.morphology import binary_dilation as imdilate
from scipy.fftpack import dct
from skimage.segmentation import find_boundaries
from pybalu.feature_extraction import lbp_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import gabor_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import haralick_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import fourier_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import hog_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import basic_int_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import basic_geo_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import fourier_des_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import hugeo_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import flusser_features # > pip3 install scipy==1.2
from pybalu.feature_extraction import gupta_features # > pip3 install scipy==1.2
from pybalu.img_processing import segbalu # > pip3 install scipy==1.2
from pyxvis.processing.images import Edge, im_grad
from pyxvis.processing.helpers.kfunctions  import gaussian_kernel
from pyxvis.processing.segmentation import seg_bimodal
from pyxvis.io.misc import dirfiles
from pyxvis.io.plots import get_image


def extract_features(fx, img=None, bw=None):
    # img grayscale image, bw binary image
    # check if I2 is a binary image (segmented region)
    if type(fx)==str:
        fx = [fx]
    n = len(fx)
    for i in range(n):
        X = extract_features_image(fx[i], img=img, bw=bw)
        if i == 0:
            Xf = X
        else:
            Xf = np.concatenate((Xf,X))
    features = np.asarray(Xf)
    return features


def extract_features_image(fxi, img=None, bw=None):
    # img grayscale image, bw binary image
    # check if I2 is a binary image (segmented region)
    if fxi == 'lbp':
        X = lbp_features(img, region=bw, hdiv=1, vdiv=1, mapping='nri_uniform')
    elif fxi == 'lbp-ri':
        X = lbp_features(img, region=bw,hdiv=1, vdiv=1, mapping='uniform')
    elif fxi == 'gabor':
        nr = 8
        nd = 8
        X = gabor_features(img, region=bw,rotations=nr, dilations=nd)
    elif fxi == 'gabor-ri':
        nr = 8
        nd = 8
        Y = gabor_features(img, region=bw,rotations=nr, dilations=nd)
        X = np.zeros((nd+3,))

        for j in range(nd):
            X[j] = np.sum(X[j*nr:(j+1)*nr-1])
        X[nr] = Y[nr*nd]
        X[nr+1] = Y[nr*nd+1]
        X[nr+2] = Y[nr*nd+2]
    elif fxi == 'hog':
        X = hog_features(img, region=bw, v_windows=1, h_windows=1, n_bins=9,
                         normalize=False, labels=False, show=False)
    elif fxi == 'haralick-1':
        X     = haralick_features(img, region=bw,distance=1)
    elif fxi == 'haralick-2':
        X     = haralick_features(img, region=bw,distance=2)
    elif fxi == 'haralick-3':
        X     = haralick_features(img, region=bw,distance=3)
    elif fxi == 'haralick-5':
        X     = haralick_features(img, region=bw,distance=5)
    elif fxi == 'haralick-7':
        X     = haralick_features(img, region=bw,distance=7)
    elif fxi == 'fourier':
        X     = fourier_features(img, region=bw)
    elif fxi == 'basicint':
        X     = basic_int_features(img, region=bw)
    elif fxi == 'clp':
        X     = clp_features(img)
    elif fxi == 'contrast':
        X     = contrast_features(img, region=bw)
    elif fxi == 'dct':
        X     = dct_features(img, region=bw)
    elif fxi == 'basicgeo':
        X = basic_geo_features(bw)
    elif fxi == 'centroid':
        f = basic_geo_features(bw)
        X = f[0:2]
    elif fxi == 'fourierdes':
        X = fourier_des_features(bw)
    elif fxi == 'flusser':
        X = flusser_features(bw)
    elif fxi == 'gupta':
        X = gupta_features(bw)
    elif fxi == 'hugeo':
        X = hugeo_features(bw)
    elif fxi == 'ellipse':
        X = ellipse_features(bw)
    else:
        print('ERROR: ' + fxi + ' does not exist as geometric feature extraction method.')
    features = np.asarray(X)
    return features


def extract_features_dir(fx, dirpath, fmt, segmentation=False):
    """

    Args:
        fx:
        dirpath:
        fmt:
        segmentation:

    Returns:

    """
    st = '*.'+fmt
    img_names = dirfiles(dirpath+'/',st)
    n = len(img_names)
    print('Extracting features in ' + str(n) + ' images...')
    for i in range(n):
        img_path = img_names[i]
        print('... reading '+dirpath+'/'+img_path)
        img = get_image(dirpath + '/' + img_path)
        gray     = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if segmentation=='balu':
            R, _, _ = segbalu(gray)
        elif segmentation=='bimodal':
            R, _ = seg_bimodal(gray)
            R = R*1
        else:
            R = None
        features = extract_features(fx,img=gray,bw=R)
        if i==0:
            m = features.shape[0]
            data = np.zeros((n,m))
            print('Extracting ' + str(features.shape[0]) + ' features...')
        data[i]    = features
    return data


def extract_features_labels(fx,dirpath,fmt,segmentation=False):
    dir_names = dirfiles(dirpath+'/','*')
    c = len(dir_names) # number of classes
    for k in range(c):
        Xk = extract_features_dir(fx,dirpath+'/'+dir_names[k],fmt,segmentation)
        dk = k*np.ones([Xk.shape[0],],dtype=int)
        if k==0:
            X = Xk
            d = dk
        else:
            X = np.concatenate((X,Xk),axis=0)
            d = np.concatenate((d,dk),axis=0)
    return X,d


def fourier_features(I,region=None,Nfourier=64,Mfourier=64,nfourier=4,mfourier=4):
    if region is None:
        region = np.ones_like(I)
    I[region == 0] = 0
    Im       = cv2.resize(I,(Nfourier,Mfourier))
    FIm      = np.fft.fft2(Im)
    Y        = FIm[0:int(Nfourier/2),0:int(Mfourier/2)]
    x        = np.abs(Y)
    F        = cv2.resize(x,(nfourier,mfourier))
    f        = np.reshape(F,(nfourier*mfourier,))
    x        = np.angle(Y)
    A        = cv2.resize(x,(nfourier,mfourier))
    a        = np.reshape(A,(nfourier*mfourier,))
    features = np.concatenate((f,a))
    return features

def dct_features(I,region=None,Ndct=64,Mdct=64,ndct=4,mdct=4):
    if region is None:
        region = np.ones_like(I)
    I[region == 0] = 0
    Im       = cv2.resize(I,(Ndct,Mdct))
    FIm      = dct(Im)
    Y        = FIm[0:int(Ndct/2),0:int(Mdct/2)]
    x        = np.abs(Y)
    F        = cv2.resize(x,(ndct,mdct))
    features = np.reshape(F,(ndct*mdct,))
    return features

def fit_ellipse(x,y):
    # Fitzgibbon, A.W., Pilu, M., and Fischer R.B., 
    # Direct least squares fitting of ellipses, 1996
    x        = x[:,None]
    y        = y[:,None]
    D        = np.hstack([x*x,x*y,y*y,x,y,np.ones(x.shape)])
    S        = np.dot(D.T,D)
    C        = np.zeros([6,6])
    C[0,2]   = C[2,0] = 2
    C[1,1]   = -1
    E,V      = np.linalg.eig(np.dot(np.linalg.inv(S),C))
    n        = np.argmax(E)
    s        = V[:,n]
    a        = s[0]
    b        = s[1]/2.
    c        = s[2]
    d        = s[3]/2.
    f        = s[4]/2.
    g        = s[5]
    dd       = b*b-a*c
    cx       = (c*d-b*f)/dd
    cy       = (a*f-b*d)/dd
    alpha    = 0.5*np.arctan(2*b/(a-c))*180/np.pi
    up       = 2*(a*f*f+c*d*d+g*b*b-2*b*d*f-a*c*g)
    down1    = (b*b-a*c)*( (c-a)*np.sqrt(1+4*b*b/((a-c)*(a-c)))-(c+a))
    down2    = (b*b-a*c)*( (a-c)*np.sqrt(1+4*b*b/((a-c)*(a-c)))-(c+a))
    a        = np.sqrt(abs(up/down1))
    b        = np.sqrt(abs(up/down2))
    area     = np.pi*a*b

    if b>a:
        ecc  = a/b
    else:
        ecc  = b/a

    features = [cx,cy,a,b,alpha,ecc,area]

    return features

def ellipse_features(R):
    E        = find_boundaries(R, mode='outer').astype(np.uint8)
    # E        = bwperim(R)
    data     = np.argwhere(E==True)
    y        = data[:,0]
    x        = data[:,1]
    features = fit_ellipse(x,y)
    return features

def clp_features(img):

    # indices of the pixels of the 8 profiles of CLP (each profile is a line of 32 pixels)
    # C[2*k,:]  : coordinate i of profile k, for k=0...7
    # C[2*k+1,:]: coordinate j of profile k, for k=0...7
    C = np.array([
        [16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [31,30,29,28,27,26,25,24,23,22,21,20,19,18,17,16,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [8,8,9,9,10,10,11,11,12,12,13,14,14,15,15,16,16,17,17,18,18,19,20,20,21,21,22,22,23,23,24,25],
        [8,8,9,9,10,10,11,11,12,12,13,14,14,15,15,16,16,17,17,18,18,19,20,20,21,21,22,22,23,23,24,25],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31],
        [25,24,23,23,22,22,21,21,20,20,19,18,18,17,17,16,16,15,15,14,14,13,12,12,11,11,10,10,9,9,8,8],
        [25,24,23,23,22,22,21,21,20,20,19,18,18,17,17,16,16,15,15,14,14,13,12,12,11,11,10,10,9,9,8,8],
        [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]])


    ng = 32
    X = np.array(np.zeros((8,ng)))
    I = cv2.resize(img, (ng, ng))
    
    for i in range(8):
        k = i*2
        for j in range(ng):
            X[i,j] = I[C[k,j],C[k+1,j]]*255
        #plt.plot(range(32),X[i,:])
    #plt.show()
    d = np.abs(X[:,0]-X[0:,31])
    k = np.argmin(d)
    y = X[k,:]
    #plt.plot(range(32),y)
    #plt.show()
    Po  = y/y[0]
    Q   = rampefree(Po)
    Qm  = np.average(Q)
    Qd  = np.max(Q)-np.min(Q)
    Qd1 = np.log(Qd+1)
    Qd2 = 2*Qd/(Po[0]+Po[ng-1])
    Qs  = np.std(Q)
    Qf  = np.abs(np.fft.fft(Q))
    Qf  = Qf[range(1,8)]
    features = [Qm, Qs, Qd, Qd1, Qd2, Qf[0], Qf[1], Qf[2], Qf[3], Qf[4], Qf[5], Qf[6]]
    return features


def rampefree(x):
    k = len(x)-1
    m = (x[k]-x[0])/k
    b = x[0]
    y = x - range(k+1)*m - b
    return y

def contrast_features(img,region=None,neihbors=2):
    img = img*255
    if region is None:
        region = np.ones_like(img)
    
    R = region==1
    Rn = R
    for i in range(neihbors):
        Rn = imdilate(Rn)

    Rn = np.bitwise_and(Rn,~R)

    if np.max(Rn)==1:
        Ir = img*region
        In = img*Rn
        MeanGr = np.average(Ir)
        MeanGn = np.average(In)
        K1 = (MeanGr-MeanGn)/MeanGn            # contrast after Kamm, 1999
        K2 = (MeanGr-MeanGn)/(MeanGr+MeanGn)   # modulation after Kamm, 1999
        K3 = np.log(MeanGr/MeanGn)                # film-contrast after Kamm, 1999
    else:
        K1 = -1        
        K2 = -1        
        K3 = -1

    (nI,mI) = img.shape

    n1 = int(round(nI/2)+1)
    m1 = int(round(mI/2)+1)

    P1 = img[n1,:]    # Profile in i-Direction
    P2 = img[:,m1]    # Profile in j-Direction
    Q1 = rampefree(P1)
    Q2 = rampefree(P2)
    Q = np.concatenate((Q1,Q2))

    Ks = np.std(Q)
    K  = np.log(np.max(Q)-np.min(Q)+1)
        
    features = [K1,K2,K3,Ks,K]
    return features








# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyxvis',
 'pyxvis.features',
 'pyxvis.geometry',
 'pyxvis.io',
 'pyxvis.learning',
 'pyxvis.processing',
 'pyxvis.processing.helpers',
 'pyxvis.simulation']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.3.2,<4.0.0',
 'numpy<1.19',
 'opencv-contrib-python>=3.4.0,<3.5.0',
 'opencv-python>=3.4.0,<3.5.0',
 'pandas>=1.1.2,<2.0.0',
 'pybalu>=0.2.9,<0.3.0',
 'pyqt5>=5.15.1,<6.0.0',
 'scikit-image>=0.17.2,<0.18.0',
 'scikit-learn>=0.23.2,<0.24.0',
 'scipy>=1.5.2,<2.0.0',
 'tensorflow>=2.3.1,<3.0.0']

setup_kwargs = {
    'name': 'pyxvis',
    'version': '0.1.0a6',
    'description': 'Python package for Xvis toolbox',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/hypermodern-python.svg)](https://pypi.org/project/hypermodern-python/)\n\n# py-XVis\n\nPython implementation for XVis Toolbox release with the book Computer Vision for X-Ray Testing. Originally implemented \nin Matlab by Domingo Mery for the first edition of the book. This package is part of the second edition of the book \nComputer Vision for X-Ray Testing (November 2020).\n\n\n# Requirements\n\n- Python 3.6 or higher\n- numpy < 1.19\n- matplotlib >= 3.3.2\n- scipy >= 1.5.2\n- pyqt5 >= 5.15.1\n- pybalu >= 0.2.9\n- opencv-python = 3.4.2\n- opencv-contrib-python = 3.4.2\n- tensorflow >= 2.3.1 \n- scikit-learn >= 0.23.2\n- scikit-image >= 0.17.2\n- pandas >= 1.1.2\n\n\n# Instalation\nThe package is part of the Python Index (PyPi). Installation is available by pip:\n\n`pip install pyxvis`\n\n\n\n# Interactive Examples\n\nAll examples in the Book have been implemented in Jupyter Notebooks tha run on Google Colab.\n\n\n## Chapter 01: X-ray Testing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1meevAFZkZ_0SUBPYqo_nl7grBcqfeVMP?usp=sharing)\n\n* Example 1.1: Displaying X-ray images\n* Example 1.2: Dual Energy\n* Example 1.3: Help of PyXvis functions\n\n\n## Chapter 02: Images for X-ray Testing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1MDc5wbH11I5PV18l5ZSt92yugz0eHi4W?usp=sharing)\n\n* Example 2.1: Displaying an X-ray image of GDXray\n\n\n## Chapter 03: Geometry in X-ray Testing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/14CgSCGcpm4Wc4SbVB70iBx6Cw6rXoZAq?usp=sharing)\n\n* Example 3.1: Euclidean 2D transformation\n* Example 3.2: Euclidean 3D transformation\n* Example 3.3: Perspective projection\n* Example 3.4: Cubic model for distortion correction\n* Example 3.5: Hyperbolic model for imaging projection\n* Example 3.6: Geometric calibration\n* Example 3.7: Epipolar geometry\n* Example 3.8: Trifocal geometry\n* Example 3.9: 3D reconstruction\n\n\n## Chapter 04: X-ray Image Processing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1edbph3qfb2PfKOWPGUWUsC0DeX8muO9U?usp=sharing)\n\n* Example 4.1: Aritmetic average of images\n* Example 4.2: Contrast enhancement\n* Example 4.3: Shading correction\n* Example 4.4: Detection of defects using median filtering\n* Example 4.5: Edge detection using gradient operation\n* Example 4.6: Edge detection with LoG\n* Example 4.7: Segmentation of bimodal images\n* Example 4.8: Welding inspection using adaptive thresholding\n* Example 4.9: Region growing\n* Example 4.10: Defects detection using LoG approach\n* Example 4.11: Segmentation using MSER\n* Example 4.12: Image restoration\n\n\n## Chapter 05: X-ray Image Representation [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1dwGTGHA1CR1om3MirGX5VCVhQgVc-g3-?usp=sharing)\n\n* Example 5.1: Geometric features\n* Example 5.2: Elliptical features\n* Example 5.3: Invariant moments\n* Example 5.4: Intenisty features\n* Example 5.5: Defect detection usin contrast features\n* Example 5.6: Crossing line profiles (CLP)\n* Example 5.7: SIFT\n* Example 5.8: feature se;ection\n* Example 5.9: Example using intenisty features\n* Example 5.10: Example using geometric features\n\n\n## Chapter 06: Classification in X-ray Testing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1zGx0HpAt7EtOiORXkTluOPDW4w5alNSj?usp=sharing)\n\n* Example 6.1: Basic classification example\n* Example 6.2: Minimal distance (dmin)\n* Example 6.3: Bayes\n* Example 6.4: Mahalanobis, LDA and QDA\n* Example 6.5: KNN\n* Example 6.6: Neural networks\n* Example 6.7: Support Vector Machines (SVM)\n* Example 6.8: Training and testing many classifiers\n* Example 6.9: Hold-out\n* Example 6.10: Cross-validation\n* Example 6.11: Confusion matrix\n* Example 6.12: ROC and Precision-Recall curves\n* Example 6.13: Example with intensity features\n* Example 6.14: Example with geometric features\n\n\n## Chapter 07: Deep Learing in X-ray Testing\n\n* Example 7.1: Basic neural networks (from skratch) [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1Ohs0hBDu5zRtNagbqBCJV6fmxq63CxS6?usp=sharing)\n\n* Example 7.2: Neural network using sklearn [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1Ohs0hBDu5zRtNagbqBCJV6fmxq63CxS6?usp=sharing)\n\n* Example 7.3: Convolutional Neural Network [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1nI3AABdBJKdT680L-ouUwX3ywpajv8bC?usp=sharing)\n\n* Example 7.4: Pre-trained models [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1JA3sgXqDHN7gkAdv1dRa-a-IgsArAA2M?usp=sharing)\n\n* Example 7.5: Fine tunning [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1iC_XLsyBru3I2RpJot8YCGt_AbQNw3mz?usp=sharing)\n\n* Example 7.6: Generative Adversarial Networks (GANs) [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1Bv9wptpLuxjXxcx6UQmPGtLdZvx949iU?usp=sharing)\n\n* Example 7.7: Object detection using YOLOv3 [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1TUBRa4kal-chsQHvstIL2ZeNZ24PJotC?usp=sharing)\n\n* Example 7.8: Object detection using YOLOv4 [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1S07jBiG1No6cq2mx8XnEV0XuToljpxs_?usp=sharing)\n\n* Example 7.9: Object detection using YOLOv5 [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1D6j2bk5uzUIJE0MQXjiyEDvh9wEEmDUJ?usp=sharing)\n\n* Example 7.10: Object detection using EfficientDet [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1EmTQ02IwXmJQ7082ooh834IYyQyxZcAL?usp=sharing)\n\n* Example 7.11: Object detection using RetinaNet [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1H7HnECaEuPIwIGWQb2vRu-eUx1LGkaa_?usp=sharing)\n\n* Example 7.12: Object detection using DETR [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1vuzCI6zE8KD3xuaS1lsCFkZKdd6NBDcY?usp=sharing)\n\n* Example 7.13: Object detection using SSD [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1PFlw9MA5z7vsUvYCA1H83bLtZRW5Zxp2?usp=sharing)\n\n\n## Chapter 08: Simulation in X-ray Testing [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1s7bKncSrQhIq_kW0qO3JvUOyyK8rfp3Q?usp=sharing)\n\n* Example 8.1: Basic simulation using voxels\n* Example 8.2: Simulation of defects using mask\n* Example 8.3: Simulation of ellipsoidal defects\n* Example 8.4: Superimposition of threat objects\n\n\n## Chapter 09: Applications in X-ray Testing\n\n* Example 9.1: Defect detection in castings [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1FLyUEYrevSu3RbZQaoPsd2BMG4MvRew0?usp=sharing)\n\n* Example 9.2: Defect detection in welds [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1mFiaoEsuhAEQoev_jgPEv35G1lIt55F8?usp=sharing)\n\n',
    'author': 'Christian Pieringer',
    'author_email': '8143906+cpieringer@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



#import matplotlib.pyplot as plt
#import matplotlib.gridspec as gridspec
#import cv2
#from PIL import Image
#import numpy as np
#from sklearn.metrics import confusion_matrix, accuracy_score
#from seaborn import heatmap
#from mpl_toolkits.mplot3d import Axes3D
import os, fnmatch
#import matplotlib.pyplot as plt
#from matplotlib.patches import Ellipse

def dirfiles(img_path,img_ext):
    img_names = fnmatch.filter(sorted(os.listdir(img_path)),img_ext)
    if '.DS_Store' in img_names:
        img_names.remove('.DS_Store')
    return img_names

def num2fixstr(x,d):
    st = '%0*d' % (d,x)
    return st




from cv2 import xfeatures2d, FlannBasedMatcher
from numpy import array


def compute_sift(img):
    sift = xfeatures2d.SIFT_create()
    keypoints, descriptors = sift.detectAndCompute(img, None)

    keypoints = array([[k.pt[0], k.pt[1], k.size, k.angle] for k in keypoints])

    return keypoints, descriptors


def compute_brief(img):
    # Initiate FAST detector
    star = xfeatures2d.StarDetector_create()

    # Initiate BRIEF extractor
    brief = xfeatures2d.BriefDescriptorExtractor_create()

    # find the keypoints with STAR
    kp = star.detect(img, None)

    # compute the descriptors with BRIEF
    keypoints, descriptors = brief.compute(img, kp)
    keypoints = array([[k.pt[0], k.pt[1], k.size, k.angle] for k in keypoints])

    return keypoints, descriptors


descriptors_funcs = {
    'sift': compute_sift,
    'orb': [],
    'brief': compute_brief,
}


def compute_descriptors(input_img, descriptor_name, **opts):
    """
    A wrapper function of different methods to compute descriptors over an image. We develop this wrapper because the
    continuous changes in the interface of the methods and the restriction in some of the descriptors available in
    OpenCV, such as, SIFT and SURF. In this way, we expose the same interface for coding.

    Args:
        input_img (ndarray):
        descriptor_name (str:
        **opts: additional options

    Returns:
        kp (ndarray): Keypoints (N, 4), where N is the number of keypoints. Colums are x, y, scale and orientation.
        desc (ndarray): Descriptors (N, D), where N is the number of detected keypoints and D the size of the descriptor.

    [1] SIFT issues in OpenCV: https://stackoverflow.com/questions/52305578/sift-cv2-xfeatures2d-sift
    -create-not-working-even-though-have-contrib-install
    """
    # Descriptor names
    if descriptor_name not in descriptors_funcs.keys():
        raise ValueError('Invalid descriptor name')

    kp, desc = descriptors_funcs[descriptor_name](input_img)

    return kp, desc


def match_descriptors(descriptors1, descriptors2, matcher='flann', max_ratio=0.8):
    """

    Args:
        descriptors1:
        descriptors2:
        matcher:
        max_ratio:

    Returns:

    [1] https://stackoverflow.com/questions/30716610/how-to-get-pixel-coordinates-from-feature-matching-in-opencv-python
    """

    if matcher is 'flann':
        # FLANN parameters
        FLANN_INDEX_KDTREE = 0

        index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
        search_params = dict(checks=50)  # or pass empty dictionary

        flann = FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(descriptors1, descriptors2, k=2)

    # Need to draw only good matches, so create a mask using ratio test as per Lowe's paper.
    good_matches = []
    for i, (m, n) in enumerate(matches):
        if m.distance < max_ratio * n.distance:
            good_matches.append([n.queryIdx, m.trainIdx])  # Keep indexes of matched keypoints
    good_matches = array(good_matches)

    return good_matches
